/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification;

import java.util.List;

import com.sigpwned.jsonification.accessor.AccessorParser;
import com.sigpwned.jsonification.accessor.AccessorProcessor;
import com.sigpwned.jsonification.accessor.JsonReference;
import com.sigpwned.jsonification.accessor.JsonReference.Multiplicity;
import com.sigpwned.jsonification.accessor.reference.IdentityJsonReference;
import com.sigpwned.jsonification.accessor.reference.ListSlotJsonReference;
import com.sigpwned.jsonification.accessor.reference.ObjectSlotJsonReference;
import com.sigpwned.jsonification.accessor.x.InvalidMultiplicityJsonAccessorEvalException;
import com.sigpwned.jsonification.model.JsonValue;

public class Accessor {
    public static Accessor parse(String accessor) {
        final Accessor[] result=new Accessor[1];
        
        AccessorParser parser=new AccessorParser(new AccessorProcessor() {
            private JsonReference reference;
            
            public void beginAccessor() {
                reference = new IdentityJsonReference();
            }

            public void endAccessor() {
                result[0] = new Accessor(reference);
            }

            public void objectAccess(List<String> names) {
                reference = new ObjectSlotJsonReference(reference, names);
            }

            public void arrayAccess(Indexes indexes) {
                reference = new ListSlotJsonReference(reference, indexes);
            }
        });
        
        parser.parse(accessor);

        return result[0];
    }
    
	public static interface Function {
		public JsonValue apply(JsonValue value);
	}
	
	private JsonReference reference;
	
	public Accessor(JsonReference reference) {
		this.reference = reference;
	}
	
	protected JsonReference getReference() {
		return reference;
	}
	
	public JsonValue readOne(JsonValue root) {
		if(getReference().getMultiplicity() != Multiplicity.ONE)
			throw new InvalidMultiplicityJsonAccessorEvalException(getReference().getMultiplicity(), getReference().toString());
		List<JsonValue> values=readMany(root);
		return values.get(0);
	}

	public JsonValue readOne(JsonValue root, JsonValue defaultValue) {
		JsonValue result;
		try {
			result = readOne(root);
		}
		catch(JsonException e) {
			result = defaultValue;
		}
		return result;
	}

	public List<JsonValue> readMany(JsonValue root) {
		return getReference().get(root);
	}
	
	public void writeOne(JsonValue root, final JsonValue set) {
		writeOne(root, new Function() {
			public JsonValue apply(JsonValue value) {
				return set;
			}
		});
	}

	public void writeOne(JsonValue root, Function transformation) {
		writeOne(root, transformation, false);
	}

	public void writeOne(JsonValue root, Function transformation, boolean failureAcceptable) {
		if(getReference().getMultiplicity() != Multiplicity.ONE)
			throw new InvalidMultiplicityJsonAccessorEvalException(getReference().getMultiplicity(), getReference().toString());
		writeMany(root, transformation, failureAcceptable);
	}

	public void writeMany(JsonValue root, Function transformation) {
		getReference().update(root, transformation);
	}

	public void writeMany(JsonValue root, Function transformation, boolean failureAcceptable) {
		getReference().update(root, transformation);
	}
}
