/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;

import com.sigpwned.jsonification.manipulation.JsonParser;
import com.sigpwned.jsonification.manipulation.parse.DefaultJsonParser;
import com.sigpwned.jsonification.manipulation.walk.DefaultJsonWalker;
import com.sigpwned.jsonification.model.JsonBoolean;
import com.sigpwned.jsonification.model.JsonFloat;
import com.sigpwned.jsonification.model.JsonInt;
import com.sigpwned.jsonification.model.JsonList;
import com.sigpwned.jsonification.model.JsonNull;
import com.sigpwned.jsonification.model.JsonObject;
import com.sigpwned.jsonification.model.JsonString;
import com.sigpwned.jsonification.model.JsonValue;
import com.sigpwned.jsonification.model.impl.JsonBooleanImpl;
import com.sigpwned.jsonification.model.impl.JsonFloatImpl;
import com.sigpwned.jsonification.model.impl.JsonIntImpl;
import com.sigpwned.jsonification.model.impl.JsonListImpl;
import com.sigpwned.jsonification.model.impl.JsonNullImpl;
import com.sigpwned.jsonification.model.impl.JsonObjectImpl;
import com.sigpwned.jsonification.model.impl.JsonStringImpl;
import com.sigpwned.jsonification.processing.EmittingJsonProcessor;
import com.sigpwned.jsonification.x.JsonParseException;

public class Json {
	public static final JsonNull NULL=JsonNullImpl.NULL;
	
	public static final JsonBoolean TRUE=JsonBooleanImpl.TRUE;
	public static final JsonBoolean FALSE=JsonBooleanImpl.FALSE;
	
	public static JsonValue parse(String json) {
		JsonValue result;
		try {
			result = parse(new StringReader(json));
		}
		catch(IOException e) {
			// We're using a StringReader, so it should never throw an
			// IOException. Just re-throw it as a parse exception if it does.
			// No need to polluate this signature with an IOException.
			throw new JsonParseException("Caught IOException: "+e.getMessage());
		}
		return result;
	}

	public static JsonValue parse(Reader input) throws IOException {
		return parse(input, JsonParser.Cleanup.defaultValue());
	}
	
	public static JsonValue parse(Reader input, JsonParser.Cleanup cleanup) throws IOException {
		return new DefaultJsonParser().parse(input, cleanup);
	}
	
	public static String emit(JsonValue json) {
		StringWriter result=new StringWriter();
		try {
			emit(json, result, EmittingJsonProcessor.Cleanup.CLOSE);
		}
		catch(IOException e) {
			throw new JsonException(e);
		}
		return result.toString();
	}
	
	public static void emit(JsonValue json, Writer out) throws IOException {
		emit(json, out, EmittingJsonProcessor.Cleanup.defaultValue());
	}
	
	public static void emit(JsonValue json, Writer out, EmittingJsonProcessor.Cleanup cleanup) throws IOException {
		new DefaultJsonWalker<Void>(new EmittingJsonProcessor(out, cleanup)).walk(json);
	}
	
	public static JsonBoolean valueOf(boolean value) {
		return value ? TRUE : FALSE;
	}

	public static JsonInt valueOf(long value) {
		return new JsonIntImpl(value);
	}

	public static JsonFloat valueOf(double value) {
		return new JsonFloatImpl(value);
	}
	
	public static JsonString valueOf(String value) {
		if(value == null)
			throw new NullPointerException();
		return new JsonStringImpl(value);
	}
	
	public static JsonList newList() {
		return new JsonListImpl();
	}
	
	public static JsonObject newObject() {
		return new JsonObjectImpl();
	}
}
