/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.accessor.reference;

import java.util.ArrayList;
import java.util.List;

import com.sigpwned.jsonification.Accessor;
import com.sigpwned.jsonification.accessor.AccessorProcessor;
import com.sigpwned.jsonification.accessor.JsonReference;
import com.sigpwned.jsonification.accessor.processing.AllIndexes;
import com.sigpwned.jsonification.accessor.processing.RangeIndexes;
import com.sigpwned.jsonification.accessor.processing.SingleIndexes;
import com.sigpwned.jsonification.accessor.x.InvalidListJsonAccessorEvalException;
import com.sigpwned.jsonification.model.JsonList;
import com.sigpwned.jsonification.model.JsonValue;
import com.sigpwned.jsonification.x.InternalErrorJsonException;

public class ListSlotJsonReference extends AbstractJsonReference {
	protected static interface IndexesWork<T> {
		public T single(SingleIndexes indexes);
		public T range(RangeIndexes indexes);
		public T all(AllIndexes indexes);
	}
	
	protected static class Interval {
		public Interval(int start, int end) {
			this.start = start;
			this.end = end;
		}
		public int start;
		public int end;
	}
	
	private JsonReference list;
	private AccessorProcessor.Indexes indexes;
	
	public ListSlotJsonReference(JsonReference list, AccessorProcessor.Indexes indexes) {
		this.list = list;
		this.indexes = indexes;
	}
	
	public List<JsonValue> get(JsonValue root) {
		List<JsonValue> result=new ArrayList<JsonValue>();
		
		List<JsonValue> containers=getList().get(root);
		for(JsonValue container : containers) {
			JsonList l=container.asList();
			Interval interval=computeInterval(l);
			for(int i=interval.start;i<interval.end;i++)
				result.add(l.get(i));
		}
		
		return result;
	}
	
	public void update(JsonValue root, Accessor.Function transformation) {
		List<JsonValue> containers=getList().get(root);
		for(JsonValue container : containers) {
			JsonList l=container.asList();
			Interval interval=computeInterval(l);
			for(int i=interval.start;i<interval.end;i++) {
				JsonValue currentValue=l.get(i);
				JsonValue newValue=transformation.apply(currentValue);
				l.set(i, newValue);
			}
		}
	}

	public JsonReference getList() {
		return list;
	}

	public AccessorProcessor.Indexes getIndexes() {
		return indexes;
	}
	
	public void append(final StringBuilder buf) {
		getList().append(buf);
		
		buf.append("[");
		withIndexes(new IndexesWork<Void>() {
			public Void single(SingleIndexes indexes) {
				buf.append(getIndexes().asSingle().getIndex());
				return null;
			}
			public Void range(RangeIndexes indexes) {
				buf.append(indexes.getStart());
				buf.append("..");
				buf.append(indexes.getEnd());
				return null;
			}
			public Void all(AllIndexes indexes) {
				buf.append("*");
				return null;
			}
		});
		buf.append("]");
	}
	
	protected Interval computeInterval(final JsonList list) {
		return withIndexes(new IndexesWork<Interval>() {
			public Interval single(SingleIndexes indexes) {
				int index=getIndexes().asSingle().getIndex();
				if(list.size() <= index)
					throw new InvalidListJsonAccessorEvalException(list, index);
				return new Interval(index, index+1);
			}
			public Interval range(RangeIndexes indexes) {
				Interval result=new Interval(getIndexes().asRange().getStart(), getIndexes().asRange().getEnd());
				if(result.start >= list.size())
					throw new InvalidListJsonAccessorEvalException(list, result.start);
				else
				if(result.start >= list.size())
					throw new InvalidListJsonAccessorEvalException(list, result.end);
				return result;
			}
			public Interval all(AllIndexes indexes) {
				return new Interval(0, list.size());
			}
		});
	}

	public Multiplicity getMultiplicity() {
		return withIndexes(new IndexesWork<Multiplicity>() {
			public Multiplicity single(SingleIndexes indexes) {
				return Multiplicity.ONE;
			}
			public Multiplicity range(RangeIndexes indexes) {
				Multiplicity result;
				if(indexes.getEnd()-indexes.getStart() == 1)
					result = Multiplicity.ONE;
				else
					result = Multiplicity.MANY;
				return result;
			}
			public Multiplicity all(AllIndexes indexes) {
				// This is slightly inaccurate. If the list this is applied to has
				// exactly one entry, this will end up ONE. But we don't know that
				// until "runtime," so just call this a MANY.
				return Multiplicity.MANY;
			}
		});
	}
	
	protected <T> T withIndexes(IndexesWork<T> work) {
		T result;
		
		if(getIndexes() instanceof SingleIndexes)
			result = work.single(indexes.asSingle());
		else
		if(getIndexes() instanceof RangeIndexes)
			result = work.range(indexes.asRange());
		else
		if(getIndexes() instanceof AllIndexes)
			result = work.all(getIndexes().asAll());
		else
			throw new InternalErrorJsonException("Unhandled indexes: "+getIndexes());
		
		return result;
	}
}