/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.manipulation;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;

import com.sigpwned.jsonification.JsonProcessor;
import com.sigpwned.jsonification.manipulation.build.JsonBuildingException;
import com.sigpwned.jsonification.manipulation.build.ProcessingJsonDocumentBuilder;
import com.sigpwned.jsonification.util.Parser;
import com.sigpwned.jsonification.x.JsonParseException;

public class JsonParser<T> extends Parser<T> {
	private JsonDocumentBuilder<T> builder;
	private StringBuilder buf;

    public JsonParser(JsonProcessor<T> processor) {
        this(new ProcessingJsonDocumentBuilder<T>(processor));
    }
    
    public JsonParser(JsonDocumentBuilder<T> builder) {
        this.builder = builder;
    }
    
	protected JsonDocumentBuilder<T> getBuilder() {
		return builder;
	}
	
	protected StringBuilder getBuf() {
		return buf;
	}
	
	public void parseFragment(InputStream input) throws IOException {
		parseFragment(new InputStreamReader(input));
	}
	
	public void parseFragment(Reader input) throws IOException {
		parseFragment(input, Cleanup.defaultValue());
	}
	
	public void parseFragment(String input) {
	    try {
	        parseFragment(new StringReader(input), Cleanup.DRAIN);
	    }
	    catch(IOException e) {
	        throw new RuntimeException(e);
	    }
	}
	
	public synchronized void parseFragment(Reader inputSource, Cleanup cleanup) throws IOException {
		setInputSource(inputSource);
		try {
			try {
				parseFragment();
			}
			finally {
				cleanup.cleanup(inputSource);
			}
		}
		finally {
			setInputSource(null);
		}
	}
	
	protected void parseFragment() throws IOException {
		this.buf = new StringBuilder();
		try {
			value(null);
		}
		finally {
			this.buf = null;
		}
	}
	
	@Override
	protected T parse() throws IOException {
		T result;
		
		JsonParseException thrown=null;
		
		try {
			parseFragment();
		}
		catch(JsonParseException e) {
		    thrown = e;
		}
		finally {
		    try {
			 result = getBuilder().end();
		    }
		    catch(JsonBuildingException e) {
		        if(thrown != null)
		            throw thrown;
		        else
		            throw e;
		    }
		}
		if(thrown != null)
		    throw thrown;
		
		return result;
	}
	
	protected void value(JsonProcessor.Coordinate coordinate) throws IOException {
		if(attempt('{')) {
		    if(coordinate != null)
		        getBuilder().beginObject(coordinate);
		    else
		        getBuilder().beginObject();
			
			do {
				whitespace();
				consume('"');
				coordinate = new JsonProcessor.Name(finishString());
				
				whitespace();
				consume(':');
				
				whitespace();
				value(coordinate);
				
				whitespace();
			} while(attempt(','));
			consume('}');
			
			getBuilder().endObject();
		} else
		if(attempt('[')) {
		    if(coordinate != null)
		        getBuilder().beginList(coordinate);
		    else
		        getBuilder().beginList();
			
			int index=0;
			do {
				whitespace();
				value(new JsonProcessor.Index(index));
				whitespace();
				index = index+1;
			} while(attempt(','));
			consume(']');
			
			getBuilder().endList();
		} else
		if(attempt('\"')) {
			String value=finishString();
            if(coordinate != null)
                getBuilder().value(coordinate, value);
            else
                getBuilder().value(value);
		} else
		if(attempti('t')) {
			consumei('r');
			consumei('u');
			consumei('e');
            if(coordinate != null)
                getBuilder().value(coordinate, true);
            else
                getBuilder().value(true);
		} else
		if(attempti('f')) {
			consumei('a');
			consumei('l');
			consumei('s');
			consumei('e');
            if(coordinate != null)
                getBuilder().value(coordinate, false);
            else
                getBuilder().value(false);
		} else
		if(attempti('n')) {
			consumei('u');
			consumei('l');
			consumei('l');
			if(coordinate != null)
			    getBuilder().nullValue(coordinate);
			else
			    getBuilder().nullValue();
		} else
		if(Character.isDigit(peek()) || peek()=='-' || peek()=='+' || peek()=='.') {
			getBuf().setLength(0);
			
			boolean floaty=false;
			for(int ch=peek();Character.isDigit(ch) || ch=='-' || ch=='+' || ch=='.' || ch=='e' || ch=='E';ch=peek()) {
				read();
				getBuf().append((char) ch);
				if(ch == '.')
					floaty = true;
				else
				if(ch=='e' || ch=='E')
					floaty = true;
			}

			if(floaty) {
				double value=Double.parseDouble(getBuf().toString());
                if(coordinate != null)
                    getBuilder().value(coordinate, value);
                else
                    getBuilder().value(value);
			}
			else {
				long value=Long.parseLong(getBuf().toString());
                if(coordinate != null)
                    getBuilder().value(coordinate, value);
                else
                    getBuilder().value(value);
			}
		}
	}
	
	private static final char[][] ESCAPES=new char[][] {
		{ 'f', '\f' },
		{ 'n', '\n' },
		{ 'r', '\r' },
		{ 't', '\t' }
	};
	protected String finishString() throws IOException {
		getBuf().setLength(0);
		while(!attempt('\"')) {
			int ch=read();
			if(ch == -1)
				 throw parseException("Unexpected EOF");
			else
			if(ch == '\\') {
				ch = read();
				if(ch == -1)
					throw parseException("Unexpected EOF");
				for(char[] escape : ESCAPES)
					if(ch == escape[0]) {
						ch = escape[1];
						break;
					}
			}
			getBuf().append((char) ch);
		}
		return getBuf().toString();
	}
}
