/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.manipulation;

import java.io.IOException;

import com.sigpwned.jsonification.Json;
import com.sigpwned.jsonification.JsonProcessor;
import com.sigpwned.jsonification.model.JsonBoolean;
import com.sigpwned.jsonification.model.JsonFloat;
import com.sigpwned.jsonification.model.JsonInt;
import com.sigpwned.jsonification.model.JsonList;
import com.sigpwned.jsonification.model.JsonNumber;
import com.sigpwned.jsonification.model.JsonObject;
import com.sigpwned.jsonification.model.JsonString;
import com.sigpwned.jsonification.model.JsonValue;
import com.sigpwned.jsonification.x.InternalErrorJsonException;

public class JsonWalker<T> {
	private JsonProcessor<T> processor;
	
	public JsonWalker(JsonProcessor<T> processor) {
		this.processor = processor;
	}
	
	protected JsonProcessor<T> getProcessor() {
		return processor;
	}
	
	public T walk(JsonValue value) throws IOException {
		T result;
		
		getProcessor().beginJson();
		try {
			walkFragment(value);
		}
		finally {
			result = getProcessor().endJson();
		}
		
		return result;
	}
	
	public void walkFragment(JsonValue value) throws IOException {
		value(null, value);
	}
	
	protected void value(JsonProcessor.Coordinate coordinate, JsonValue value) throws IOException {
		if(value == Json.NULL)
			getProcessor().nullValue(coordinate);
		else
		if(value instanceof JsonBoolean)
			getProcessor().value(coordinate, value.asBoolean().booleanVal());
		else
		if(value instanceof JsonNumber) {
			JsonNumber number=value.asNumber();
			if(number.getOriginalType() == JsonNumber.OriginalType.FLOAT)
				getProcessor().value(coordinate, number.doubleVal());
			else
			if(number.getOriginalType() == JsonNumber.OriginalType.INT)
				getProcessor().value(coordinate, number.longVal());
			else
				throw new InternalErrorJsonException("Unhandled OriginalType: "+number.getOriginalType());
		} else
		if(value instanceof JsonFloat)
			getProcessor().value(coordinate, value.asFloat().doubleVal());
		else
		if(value instanceof JsonInt)
			getProcessor().value(coordinate, value.asInt().longVal());
		else
		if(value instanceof JsonList) {
			getProcessor().beginList(coordinate);
			try {
				int index=0;
				for(JsonValue element : value.asList()) {
					value(new JsonProcessor.Index(index), element);
					index = index+1;
				}
			}
			finally {
				getProcessor().endList();
			}
		} else
		if(value instanceof JsonObject) {
			getProcessor().beginObject(coordinate);
			try {
				JsonObject object=value.asObject();
				for(String key : object.keySet())
					value(new JsonProcessor.Name(key), object.get(key));
			}
			finally {
				getProcessor().endObject();
			}
		} else
		if(value instanceof JsonString)
			getProcessor().value(coordinate, value.asString().stringVal());
		else
			throw new InternalErrorJsonException("Unhandled value: "+value);
	}
}
