/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.manipulation.build;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.sigpwned.jsonification.JsonProcessor;
import com.sigpwned.jsonification.JsonProcessor.Coordinate;
import com.sigpwned.jsonification.manipulation.JsonFragmentBuilder;

public class BufferedJsonFragmentBuilder extends AbstractJsonFragmentBuilder {
	protected static interface Event {
		public void replay(JsonFragmentBuilder builder) throws IOException;
	}

	private List<Event> events;
	
	public BufferedJsonFragmentBuilder() {
		this.events = new ArrayList<Event>();
	}
	
	protected List<Event> getEvents() {
		if(events == null)
			throw new IllegalStateException("Already replayed");
		return events;
	}
	
	protected void setEvents(List<Event> events) {
		this.events = events;
	}
	
	protected void addEvent(Event event) {
		getEvents().add(event);
	}
	
	public void replay(JsonFragmentBuilder other) throws IOException {
		for(Event event : getEvents())
			event.replay(other);
		setEvents(null);
	}
	
	public BufferedJsonFragmentBuilder name(final String name) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.name(name);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder beginObject() throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.beginObject();
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder beginObject(final String name) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.beginObject(name);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder beginObject(final Coordinate coordinate) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.beginObject(coordinate);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder endObject() throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.endObject();
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder beginList() throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.beginList();
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder beginList(final String name) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.beginList(name);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder beginList(final Coordinate coordinate) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.beginList(coordinate);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder endList() throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.endList();
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final long value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final String name, final long value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(name, value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final JsonProcessor.Coordinate coordinate, final long value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(coordinate, value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final double value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final String name, final double value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(name, value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final JsonProcessor.Coordinate coordinate, final double value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(coordinate, value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final String value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final String name, final String value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(name, value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final JsonProcessor.Coordinate coordinate, final String value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(coordinate, value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final boolean value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final String name, final boolean value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(name, value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder value(final JsonProcessor.Coordinate coordinate, final boolean value) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.value(coordinate, value);
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder nullValue() throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.nullValue();
			}
		});
		return this;
	}
	
	public BufferedJsonFragmentBuilder nullValue(final String name) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.nullValue(name);
			}
		});
		return this;
	}

	public BufferedJsonFragmentBuilder nullValue(final JsonProcessor.Coordinate coordinate) throws IOException {
		addEvent(new Event() {
			public void replay(JsonFragmentBuilder builder) throws IOException {
				builder.nullValue(coordinate);
			}
		});
		return this;
	}
}
