/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.manipulation.build;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.sigpwned.jsonification.JsonProcessor;
import com.sigpwned.jsonification.JsonProcessor.Coordinate;
import com.sigpwned.jsonification.manipulation.JsonDocumentBuilder;
import com.sigpwned.jsonification.manipulation.build.processing.CoordinateFactory;
import com.sigpwned.jsonification.manipulation.build.processing.ScopeType;
import com.sigpwned.jsonification.manipulation.build.processing.State;
import com.sigpwned.jsonification.manipulation.build.processing.coordinates.ListCoordinateFactory;
import com.sigpwned.jsonification.manipulation.build.processing.coordinates.ObjectCoordinateFactory;
import com.sigpwned.jsonification.manipulation.build.processing.coordinates.TopCoordinateFactory;
import com.sigpwned.jsonification.manipulation.build.x.IllegalEventJsonBuildingException;
import com.sigpwned.jsonification.x.InternalErrorJsonException;

public class ProcessingJsonDocumentBuilder<T> extends AbstractJsonDocumentBuilder<T> {
    public static interface InitialContext {
    }
    
    public static class ListInitialContext implements InitialContext {
        private int nextIndex;
        
        public ListInitialContext(int nextIndex) {
            this.nextIndex = nextIndex;
        }
        
        public int getNextIndex() {
            return nextIndex;
        }
    }
    
    public static class ObjectInitialContext implements InitialContext {
        private String nextName;
        
        public ObjectInitialContext(String nextName) {
            this.nextName = nextName;
        }
        
        public String getNextName() {
            return nextName;
        }
    }
    
    /**
     * It may be desirable for this JsonDocumentBuilder to function as a
     * JsonFragmentBuilder at times. To this end, we allow the user to put it
     * into a FRAGMENT mode, which means it will never call beginJson() or
     * endJson() on its internal JsonProcessor.
     */
    protected static enum Mode {
        DOCUMENT {
            @Override
            public State complete() {
                return State.DONE;
            }

            @Override
            public State initialState() {
                return State.UNINITIALIZED;
            }

            @Override
            public void initializeCoordinateFactory(ProcessingJsonDocumentBuilder<?> builder, InitialContext initialContext) {
                if(initialContext != null)
                    throw new IllegalArgumentException("Cannot specify InitialContext in document mode");
                builder.push(new TopCoordinateFactory());
            }
        },
        FRAGMENT {
            @Override
            public State complete() {
                throw new IllegalStateException("Cannot end() in Fragment mode");
            }

            @Override
            public State initialState() {
                return State.READY;
            }

            @Override
            public void initializeCoordinateFactory(ProcessingJsonDocumentBuilder<?> builder, InitialContext initialContext) {
                if(initialContext != null) {
                    // This means we're starting somewhere, like in a list or
                    // object. This primes our builder.
                    if(initialContext instanceof ObjectInitialContext) {
                        ObjectInitialContext o=(ObjectInitialContext) initialContext;
                        CoordinateFactory f=new ObjectCoordinateFactory(false);
                        if(o.getNextName() != null)
                            f.primeCoordinate(o.getNextName());
                        builder.push(f);
                    } else
                    if(initialContext instanceof ListInitialContext) {
                        ListInitialContext l=(ListInitialContext) initialContext;
                        CoordinateFactory f=new ListCoordinateFactory(false, l.getNextIndex());
                        builder.push(f);
                    }
                    else
                        throw new InternalErrorJsonException("Unhandled InitialContext: "+initialContext);
                }
                else {
                    // This means we're starting at document root. We still
                    // don't want to call beginJson() or endJson(), though.
                    builder.push(new TopCoordinateFactory());
                }
            }
        };
        
        public abstract State initialState();
        
        public abstract State complete();
        
        public abstract void initializeCoordinateFactory(ProcessingJsonDocumentBuilder<?> builder, InitialContext initialContext);
    }
    
    private Mode mode;
	private JsonProcessor<T> processor;
	private State state;
	private List<CoordinateFactory> coordinateFactories;
	
    /**
     * Creates a new ProcessingJsonDocumentBuilder in "document mode." Said
     * differently, this builder assumes it is starting at the "root" of the
     * JSON document being built, and will call beginJson()/endJson() on its
     * child processor.
     */
    public ProcessingJsonDocumentBuilder(JsonProcessor<T> processor) {
        this(Mode.DOCUMENT, processor, null);
    }
    
    /**
     * Creates a new ProcessingJsonDocumentBuilder in "fragment mode." Said
     * differently, this builder assumes it is NOT at the "root" of the JSON
     * document being built, and so will not call beginJson()/endJson() on its
     * child processor. (Calling end() on this builder will result in an
     * Exception.) Additionally, the builder assumes its current state -- in
     * a List at position 3, in an Object with next value named "foo", etc --
     * is as indicated by its InitialContext parameter. Attempting to descend
     * "below" the initial state of this builder using endObject() or
     * endList() will also result in an Exception. As always, begin/end calls
     * must be paired appropriately. (You cannot begin with a ListInitialContext
     * and call endList() immediately, for example.)
     */
    public ProcessingJsonDocumentBuilder(JsonProcessor<T> processor, InitialContext initialContext) {
        this(Mode.FRAGMENT, processor, initialContext);
    }
    
    protected ProcessingJsonDocumentBuilder(Mode mode, JsonProcessor<T> processor, InitialContext context) {
        this.mode = mode;
        this.processor = processor;
        this.state = mode.initialState();
        this.coordinateFactories = new ArrayList<CoordinateFactory>();
        this.mode.initializeCoordinateFactory(this, context);
    }
    
	public JsonProcessor<T> getProcessor() {
		return processor;
	}
	
	protected String currentPath() {
		StringBuilder result=new StringBuilder();

		if(getCoordinateFactories().size() == 0) {
			// This is OK. An empty path is what we want, and we've got it.
		}
		else {
			result.append(getCoordinateFactories().get(0).currentCoordinate());
			for(int i=1;i<getCoordinateFactories().size()-1;i++)
				result.append(getCoordinateFactories().get(i).currentCoordinate());
			if(getCoordinateFactories().get(getCoordinateFactories().size()-1).currentCoordinate() != null)
				result.append(getCoordinateFactories().get(getCoordinateFactories().size()-1).currentCoordinate());
		}
		
		return result.toString();
	}
	
	protected void append(StringBuilder buf, JsonProcessor.Coordinate coordinate) {
		if(coordinate == null) {
			// Every CoordinateFactory until the last one ought to have a scope
			// set. We don't validate that here. (If the user wants that
			// validated, they should use a ValidatingJsonProcessor.) So, if
			// that's not the case here, we spit out a "(?)" to note bad scopes.
			buf.append("(?)");
		} else
		if(coordinate.getName() != null) {
			buf.append(".");
			buf.append(coordinate.getName());
		} else
		if(coordinate.getIndex() != null) {
			buf.append("[");
			buf.append(coordinate.getIndex().intValue());
			buf.append("]");
		}
		else
			throw new InternalErrorJsonException("Unhandled coordinate: "+coordinate);
	}
	
	protected List<CoordinateFactory> getCoordinateFactories() {
		return coordinateFactories;
	}
	
	protected void push(CoordinateFactory coordinatefactory) {
		getCoordinateFactories().add(coordinatefactory);
	}
	
	protected CoordinateFactory top(ScopeType scopeType) {
		CoordinateFactory result=getCoordinateFactories().get(getCoordinateFactories().size()-1);
		if(!scopeType.accepts(result.getScopeType()))
			throw new IllegalEventJsonBuildingException("Expected `"+scopeType.getScopeTypeName()+"' scope, found `"+result.getScopeTypeName()+"' (path=`"+currentPath()+"')");
		return result;
	}
	
	protected CoordinateFactory pop(ScopeType scopeType) {
		CoordinateFactory result=top(scopeType);
		if(!result.isPoppable())
	        throw new IllegalEventJsonBuildingException("No open scopes; top="+result);
		getCoordinateFactories().remove(getCoordinateFactories().size()-1);
		return result;
	}
	
	protected void step() throws IOException {
		state = state.step(this);
	}
	
	public T end() throws IOException {
        State nextState=mode.complete();
        T result=state.complete(this);
        state = nextState;
        return result;
	}
	
	public JsonDocumentBuilder<T> name(String name) throws IOException {
		top(ScopeType.ANY).primeCoordinate(name);
		return this;
	}
	
	public JsonDocumentBuilder<T> beginObject() throws IOException {
		return beginObject(top(ScopeType.ANY).nextCoordinate(null));
	}
	
	public JsonDocumentBuilder<T> beginObject(String name) throws IOException {
		return beginObject(top(ScopeType.OBJECT).nextCoordinate(name));
	}
	
	public JsonDocumentBuilder<T> beginObject(Coordinate coordinate) throws IOException {
		step();
		push(new ObjectCoordinateFactory(true));
		getProcessor().beginObject(coordinate);
		return this;
	}
	
	public JsonDocumentBuilder<T> endObject() throws IOException {
		step();
		getProcessor().endObject();
		pop(ScopeType.OBJECT);
		return this;
	}
	
	public JsonDocumentBuilder<T> beginList() throws IOException {
		return beginList(top(ScopeType.ANY).nextCoordinate(null));
	}
	
	public JsonDocumentBuilder<T> beginList(String name) throws IOException {
		return beginList(top(ScopeType.OBJECT).nextCoordinate(name));
	}
	
	public JsonDocumentBuilder<T> beginList(JsonProcessor.Coordinate coordinate) throws IOException {
		step();
		push(new ListCoordinateFactory(true));
		getProcessor().beginList(coordinate);
		return this;
	}
	
	public JsonDocumentBuilder<T> endList() throws IOException {
		step();
		getProcessor().endList();
		pop(ScopeType.LIST);
		return this;
	}
	
	public JsonDocumentBuilder<T> value(long value) throws IOException {
		return value(top(ScopeType.ANY).nextCoordinate(null), value);
	}
	
	public JsonDocumentBuilder<T> value(String name, long value) throws IOException {
		return value(top(ScopeType.OBJECT).nextCoordinate(name), value);
	}
	
	public JsonDocumentBuilder<T> value(JsonProcessor.Coordinate coordinate, long value) throws IOException {
		step();
		getProcessor().value(coordinate, value);
		return this;
	}
	
	public JsonDocumentBuilder<T> value(double value) throws IOException {
		return value(top(ScopeType.ANY).nextCoordinate(null), value);
	}
	
	public JsonDocumentBuilder<T> value(String name, double value) throws IOException {
		return value(top(ScopeType.OBJECT).nextCoordinate(name), value);
	}
	
	public JsonDocumentBuilder<T> value(JsonProcessor.Coordinate coordinate, double value) throws IOException {
		step();
		getProcessor().value(coordinate, value);
		return this;
	}
	
	public JsonDocumentBuilder<T> value(String value) throws IOException {
		return value(top(ScopeType.ANY).nextCoordinate(null), value);
	}
	
	public JsonDocumentBuilder<T> value(String name, String value) throws IOException {
		return value(top(ScopeType.OBJECT).nextCoordinate(name), value);
	}
	
	public JsonDocumentBuilder<T> value(JsonProcessor.Coordinate coordinate, String value) throws IOException {
		step();
		getProcessor().value(coordinate, value);
		return this;
	}
	
	public JsonDocumentBuilder<T> value(boolean value) throws IOException {
		return value(top(ScopeType.ANY).nextCoordinate(null), value);
	}
	
	public JsonDocumentBuilder<T> value(String name, boolean value) throws IOException {
		return value(top(ScopeType.OBJECT).nextCoordinate(name), value);
	}
	
	public JsonDocumentBuilder<T> value(JsonProcessor.Coordinate coordinate, boolean value) throws IOException {
		step();
		getProcessor().value(coordinate, value);
		return this;
	}
	
	public JsonDocumentBuilder<T> nullValue() throws IOException {
		return nullValue(top(ScopeType.ANY).nextCoordinate(null));
	}
	
	public JsonDocumentBuilder<T> nullValue(String name) throws IOException {
		return nullValue(top(ScopeType.OBJECT).nextCoordinate(name));
	}

	public JsonDocumentBuilder<T> nullValue(Coordinate coordinate) throws IOException {
		step();
		getProcessor().nullValue(coordinate);
		return this;
	}
}
