/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.processing;

import java.util.ArrayList;
import java.util.List;

import com.sigpwned.jsonification.Json;
import com.sigpwned.jsonification.JsonProcessor;
import com.sigpwned.jsonification.manipulation.build.JsonBuildingException;
import com.sigpwned.jsonification.model.JsonBoolean;
import com.sigpwned.jsonification.model.JsonFloat;
import com.sigpwned.jsonification.model.JsonInt;
import com.sigpwned.jsonification.model.JsonList;
import com.sigpwned.jsonification.model.JsonNull;
import com.sigpwned.jsonification.model.JsonObject;
import com.sigpwned.jsonification.model.JsonString;
import com.sigpwned.jsonification.model.JsonValue;
import com.sigpwned.jsonification.util.AbstractJsonProcessor;

public class BuildingJsonProcessor extends AbstractJsonProcessor<JsonValue> {
	private List<JsonValue> values;
	private JsonValue result;
	
	public BuildingJsonProcessor() {
	}
	
	public void beginJson() {
		values = new ArrayList<JsonValue>();
	}

	public JsonValue endJson() {
        if(!empty())
            throw new JsonBuildingException("Unbalanced begin()/end() calls; top="+top());
		values = null;
		
		JsonValue result=this.result;
		this.result = null;
		
		return result;
	}

	public void beginObject(JsonProcessor.Coordinate coordinate) {
		JsonObject value=Json.newObject();
		if(coordinate!=null && coordinate.getName()!=null)
			top().asObject().put(coordinate.getName(), value);
		push(value);
	}

	public void endObject() {
		pop().asObject();
	}

	public void beginList(JsonProcessor.Coordinate coordinate) {
		JsonList value=Json.newList();
		if(coordinate != null) {
			if(coordinate.getName() != null)
				top().asObject().put(coordinate.getName(), value);
			else
			if(coordinate.getIndex() != null)
				top().asList().add(value);
		}
		push(value);
	}

	public void endList() {
		pop().asList();
	}

	public void value(JsonProcessor.Coordinate coordinate, boolean v) {
		JsonBoolean value=Json.valueOf(v);
		if(coordinate != null) {
			if(coordinate.getName() != null)
				top().asObject().put(coordinate.getName(), value);
			else
			if(coordinate.getIndex() != null)
				top().asList().add(value);
		}
		value(value);
	}

	public void value(JsonProcessor.Coordinate coordinate, long v) {
		JsonInt value=Json.valueOf(v);
		if(coordinate != null) {
			if(coordinate.getName() != null)
				top().asObject().put(coordinate.getName(), value);
			else
			if(coordinate.getIndex() != null)
				top().asList().add(value);
		}
		value(value);
	}

	public void value(JsonProcessor.Coordinate coordinate, double v) {
		JsonFloat value=Json.valueOf(v);
		if(coordinate != null) {
			if(coordinate.getName() != null)
				top().asObject().put(coordinate.getName(), value);
			else
			if(coordinate.getIndex() != null)
				top().asList().add(value);
		}
		value(value);
	}

	public void value(JsonProcessor.Coordinate coordinate, String v) {
		if(v == null)
			throw new NullPointerException();
		JsonString value=Json.valueOf(v);
		if(coordinate != null) {
			if(coordinate.getName() != null)
				top().asObject().put(coordinate.getName(), value);
			else
			if(coordinate.getIndex() != null)
				top().asList().add(value);
		}
		value(value);
	}

	public void nullValue(JsonProcessor.Coordinate coordinate) {
		JsonNull value=Json.NULL;
		if(coordinate != null) {
			if(coordinate.getName() != null)
				top().asObject().put(coordinate.getName(), value);
			else
			if(coordinate.getIndex() != null)
				top().asList().add(value);
		}
		value(value);
	}
	
	protected List<JsonValue> getValues() {
		if(values == null)
			throw new IllegalStateException("Not parsing");
		return values;
	}

	protected JsonValue top() {
		if(getValues().size() == 0)
			throw new IllegalStateException("Empty");
		return getValues().get(getValues().size()-1);
	}
	
	protected JsonValue pop() {
		JsonValue result=top();
		getValues().remove(getValues().size()-1);
		return result;
	}
	
	protected void push(JsonValue value) {
		if(empty()) {
            if(result != null)
                throw new JsonBuildingException("Cannot parse one value after another at the top scope");
            else
                result = value;
		}
		getValues().add(value);
	}
	
	protected void value(JsonValue value) {
	    if(empty()) {
	        if(result != null)
	            throw new JsonBuildingException("Cannot parse one value after another at the top scope");
	        else
	            result = value;
	    }
	}
	
	protected boolean empty() {
		return getValues().isEmpty();
	}

	protected int size() {
		return getValues().size();
	}
}
