/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.processing;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.List;

import com.sigpwned.jsonification.util.AbstractJsonProcessor;
import com.sigpwned.jsonification.x.JsonEmitException;

public class EmittingJsonProcessor extends AbstractJsonProcessor<Void> {
	public static enum Cleanup {
		NOP {
			@Override
			public void cleanUp(Writer outputDrain) {
			}
		},
		FLUSH {
			@Override
			public void cleanUp(Writer outputDrain) throws IOException {
				outputDrain.flush();
			}
		},
		CLOSE {
			@Override
			public void cleanUp(Writer outputDrain) throws IOException {
				outputDrain.close();
			}
		};
		
		public static Cleanup defaultValue() {
			return CLOSE;
		}
		
		public abstract void cleanUp(Writer outputDrain) throws IOException;
	}
	
	private Writer outputDrain;
	private Cleanup cleanup;
	private List<Integer> elements;
	
	public EmittingJsonProcessor(OutputStream output) {
		this(new OutputStreamWriter(output), Cleanup.defaultValue());
	}
	
	public EmittingJsonProcessor(Writer output) {
		this(output, Cleanup.defaultValue());
	}
	
	public EmittingJsonProcessor(Writer outputDrain, Cleanup cleanup) {
		if(outputDrain == null)
			throw new NullPointerException();
		if(cleanup == null)
			throw new NullPointerException();
		this.outputDrain = outputDrain;
		this.cleanup = cleanup;
		this.elements = new ArrayList<Integer>();
	}
	
	protected Writer getOutputDrain() {
		if(outputDrain == null)
			throw new IllegalStateException("Already cleaned up");
		return outputDrain;
	}
	
	protected Cleanup getCleanup() {
		return cleanup;
	}
	
	protected void write(String s) throws IOException {
		getOutputDrain().write(s);
	}

	protected void writeEscaped(String s) throws IOException {
		int index=0;
		for(int pos=s.indexOf('"', index);pos!=-1;pos=s.indexOf('"', index)) {
			getOutputDrain().write(s, index, pos-index);
			getOutputDrain().write("\\\"");
			index = pos+1;
		}
		getOutputDrain().write(s, index, s.length());
	}
	
	protected List<Integer> getElements() {
		return elements;
	}
	
	protected void push(Integer element) {
		getElements().add(element);
	}
	
	protected Integer top() {
		if(getElements().size() == 0)
			throw new JsonEmitException("Not in any scopes");
		return getElements().get(getElements().size()-1);
	}
	
	protected void top(Integer element) {
		if(getElements().size() == 0)
			throw new JsonEmitException("Not in any scopes");
		getElements().set(getElements().size()-1, element);
	}
	
	protected Integer pop() {
		Integer result=top();
		getElements().remove(getElements().size()-1);
		return result;
	}
	
	protected void separator() throws IOException {
		if(top()!=null) {
			if(top().intValue()!=0)
				write(",");
			top(Integer.valueOf(top().intValue()+1));
		}
		
	}
	
	@Override
	public void beginJson() {
		push(null);
	}

	@Override
	public Void endJson() throws IOException {
		if(pop() != null)
			throw new JsonEmitException("Scopes still open");
		if(getElements().size() != 0)
			throw new JsonEmitException("Scopes still open");
		try {
			getCleanup().cleanUp(outputDrain);
		}
		finally {
			outputDrain = null;
		}
		return null;
	}
	
	protected void coordinate(Coordinate coordinate) throws IOException {
		if(coordinate != null)
			if(coordinate.getName() != null) {
				write("\"");
				writeEscaped(coordinate.getName());
				write("\":");
			}
	}

	@Override
	public void beginObject(Coordinate coordinate) throws IOException {
		separator();
		coordinate(coordinate);
		push(Integer.valueOf(0));
		write("{");
	}

	@Override
	public void endObject() throws IOException {
		write("}");
		if(pop() == null)
			throw new JsonEmitException("Scope begin/end mismatch");
	}

	@Override
	public void beginList(Coordinate coordinate) throws IOException {
		separator();
		coordinate(coordinate);
		write("[");
		push(Integer.valueOf(0));
	}

	@Override
	public void endList() throws IOException {
		write("]");
		if(pop() == null)
			throw new JsonEmitException("Scope begin/end mismatch");
	}

	@Override
	public void value(Coordinate coordinate, boolean value) throws IOException {
		separator();
		coordinate(coordinate);
		write(Boolean.toString(value));
	}

	@Override
	public void value(Coordinate coordinate, long value) throws IOException {
		separator();
		coordinate(coordinate);
		write(Long.toString(value));
	}

	@Override
	public void value(Coordinate coordinate, double value) throws IOException {
		separator();
		coordinate(coordinate);
		write(Double.toString(value));
	}

	@Override
	public void value(Coordinate coordinate, String value) throws IOException {
		separator();
		coordinate(coordinate);
		write("\"");
		writeEscaped(value);
		write("\"");
	}

	@Override
	public void nullValue(Coordinate coordinate) throws IOException {
		separator();
		coordinate(coordinate);
		write("null");
	}
}