/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.processing;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.sigpwned.jsonification.JsonProcessor;
import com.sigpwned.jsonification.util.AbstractJsonProcessor;
import com.sigpwned.jsonification.x.InternalErrorJsonException;
import com.sigpwned.jsonification.x.JsonValidationException;

public class ValidatingJsonProcessor<T> extends AbstractJsonProcessor<T> {
	protected static enum State {
		UNINITIALIZED {
			@Override
			public State beginJson() {
				return PROCESSING;
			}

			@Override
			public State value() {
				throw new JsonValidationException("Must call beginJson() before emitting a value");
			}

			@Override
			public State endJson() {
				throw new JsonValidationException("Must call beginJson() before endJson()");
			}
		},
		PROCESSING {
			@Override
			public State beginJson() {
				throw new JsonValidationException("Already called beginJson()");
			}

			@Override
			public State value() {
				return PROCESSING;
			}

			@Override
			public State endJson() {
				return COMPLETE;
			}
		},
		COMPLETE {
			@Override
			public State beginJson() {
				throw new JsonValidationException("Already called endJson()");
			}

			@Override
			public State value() {
				throw new JsonValidationException("Already called endJson()");
			}

			@Override
			public State endJson() {
				throw new JsonValidationException("Already called endJson()");
			}
		};
		
		public abstract State beginJson();
		
		public abstract State value();
		
		public abstract State endJson();
	}
	
	protected static enum ScopeType {
		TOP {
			@Override
			public JsonProcessor.Coordinate coordinate(JsonProcessor.Coordinate suggestion, JsonProcessor.Coordinate coordinate) {
				if(coordinate != null)
					throw new JsonValidationException("Cannot specify coordinate on top scope");
				return null;
			}

			@Override
			public com.sigpwned.jsonification.JsonProcessor.Coordinate initialSuggestion() {
				return null;
			}
		},
		OBJECT {
			@Override
			public JsonProcessor.Coordinate coordinate(JsonProcessor.Coordinate suggestion, JsonProcessor.Coordinate coordinate) {
				if(coordinate == null)
					throw new JsonValidationException("Must specify name for values inside a JSON object");
				else
				if(coordinate.getIndex() != null)
					throw new JsonValidationException("Must specify name (not index) for values inside a JSON object; index="+coordinate.getIndex());
				else
				if(coordinate.getName() == null) {
					// Internal error
					throw new InternalErrorJsonException("Name unexpectedly null");
				}
				return coordinate;
			}

			@Override
			public com.sigpwned.jsonification.JsonProcessor.Coordinate initialSuggestion() {
				return null;
			}
		},
		LIST {
			@Override
			public JsonProcessor.Coordinate coordinate(JsonProcessor.Coordinate suggestion, JsonProcessor.Coordinate coordinate) {
				if(coordinate == null)
					throw new JsonValidationException("Must specify index for values inside a JSON object");
				else
				if(coordinate.getName() != null)
					throw new JsonValidationException("Must specify index (not name) for values inside a JSON object; index="+coordinate.getName());
				else
				if(coordinate.getIndex() == null) {
					// Internal error
					throw new InternalErrorJsonException("Index unexpectedly null");
				} else
				if(coordinate.getIndex() != suggestion.getIndex()+1)
					throw new JsonValidationException("Expected list index "+(suggestion.getIndex()+1)+", got index "+coordinate.getIndex());
				return coordinate;
			}

			@Override
			public com.sigpwned.jsonification.JsonProcessor.Coordinate initialSuggestion() {
				return new JsonProcessor.Index(-1);
			}
		};
		
		public abstract JsonProcessor.Coordinate initialSuggestion();
		
		public abstract JsonProcessor.Coordinate coordinate(JsonProcessor.Coordinate suggestion, JsonProcessor.Coordinate coordinate);
	}
	
	protected static class Scope {
		public Scope(ScopeType type) {
			this.type = type;
			this.suggestedCoordinate = type.initialSuggestion();
		}
		
		private ScopeType type;
		private JsonProcessor.Coordinate suggestedCoordinate;
		
		public JsonProcessor.Coordinate coordinate(JsonProcessor.Coordinate coordinate) {
			setSuggestedCoordinate(getType().coordinate(getSuggestedCoordinate(), coordinate));
			return getSuggestedCoordinate();
		}
		
		public ScopeType getType() {
			return type;
		}
		
		protected void setSuggestedCoordinate(JsonProcessor.Coordinate suggestedCoordinate) {
			this.suggestedCoordinate = suggestedCoordinate;
		}
		
		public JsonProcessor.Coordinate getSuggestedCoordinate() {
			return suggestedCoordinate;
		}
	}
	
	private JsonProcessor<T> inner;
	private State state;
	private List<Scope> scopes;
	
	public ValidatingJsonProcessor(JsonProcessor<T> inner) {
		this.inner = inner;
		this.state = State.UNINITIALIZED;
		this.scopes = new ArrayList<Scope>();
	}
	
	protected JsonProcessor<T> getInner() {
		return inner;
	}
	
	protected State getState() {
		return state;
	}
	
	protected List<Scope> getScopes() {
		return scopes;
	}
	
	protected void push(Scope scope) {
		if(getScopes().size()!=0 && scope.type==ScopeType.TOP)
			throw new InternalErrorJsonException("Cannot add Top Scope");
		getScopes().add(scope);
	}
	
	protected Scope top() {
		if(getScopes().size() == 0)
			throw new JsonValidationException("Haven't called beginJson() yet");
		return getScopes().get(getScopes().size()-1);
	}
	
	protected Scope pop(ScopeType scopeType) {
		Scope result=top();
		if(result.type != scopeType)
			throw new JsonValidationException("Not in a "+scopeType+" type");
		getScopes().remove(getScopes().size()-1);
		return result;
	}

	@Override
	public void beginJson() throws IOException {
		state = state.beginJson();
		push(new Scope(ScopeType.TOP));
		getInner().beginJson();
	}

	@Override
	public T endJson() throws IOException {
		state = state.endJson();
		pop(ScopeType.TOP);
		
		if(getScopes().size() != 0)
			throw new InternalErrorJsonException("Scopes unexpectedly not empty (scopes="+getScopes()+")");
		
		if(state != State.COMPLETE)
			throw new InternalErrorJsonException("Unexpected state (state="+state+")");
		
		return getInner().endJson();
	}

	@Override
	public void beginObject(JsonProcessor.Coordinate coordinate) throws IOException {
		state      = state.value();
		coordinate = top().coordinate(coordinate);
		getInner().beginObject(coordinate);
		push(new Scope(ScopeType.OBJECT));
	}

	@Override
	public void endObject() throws IOException {
		pop(ScopeType.OBJECT);
		getInner().endObject();
	}

	@Override
	public void beginList(JsonProcessor.Coordinate coordinate) throws IOException {
		state      = state.value();
		coordinate = top().coordinate(coordinate);
		getInner().beginList(coordinate);
		push(new Scope(ScopeType.LIST));
	}

	@Override
	public void endList() throws IOException {
		pop(ScopeType.LIST);
		getInner().endList();
	}

	@Override
	public void value(JsonProcessor.Coordinate coordinate, boolean value) throws IOException {
		state      = state.value();
		coordinate = top().coordinate(coordinate);
		getInner().value(coordinate, value);
	}

	@Override
	public void value(JsonProcessor.Coordinate coordinate, long value) throws IOException {
		state      = state.value();
		coordinate = top().coordinate(coordinate);
		getInner().value(coordinate, value);
	}

	@Override
	public void value(JsonProcessor.Coordinate coordinate, double value) throws IOException {
		state      = state.value();
		coordinate = top().coordinate(coordinate);
		getInner().value(coordinate, value);
	}

	@Override
	public void value(JsonProcessor.Coordinate coordinate, String value) throws IOException {
		state      = state.value();
		coordinate = top().coordinate(coordinate);
		getInner().value(coordinate, value);
	}

	@Override
	public void nullValue(JsonProcessor.Coordinate coordinate) throws IOException {
		state      = state.value();
		coordinate = top().coordinate(coordinate);
		getInner().nullValue(coordinate);
	}
}
