/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.serialization;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.sigpwned.jsonification.JsonSerializer;
import com.sigpwned.jsonification.manipulation.JsonDocumentBuilder;
import com.sigpwned.jsonification.manipulation.JsonFragmentBuilder;
import com.sigpwned.jsonification.serialization.serialize.JsonSerializeApproach;
import com.sigpwned.jsonification.serialization.serialize.JsonTypeSerializer;

public abstract class AbstractJsonSerializer implements JsonSerializer {
	private Map<Class<?>,JsonTypeSerializer<?>> typeSerializers;
	private List<JsonSerializeApproach> serializeApproaches;
	private ExceptionPolicy exceptionPolicy;
	
	public AbstractJsonSerializer() {
		this.typeSerializers = new HashMap<Class<?>,JsonTypeSerializer<?>>();
		this.serializeApproaches = new ArrayList<JsonSerializeApproach>();
	}

	protected Map<Class<?>, JsonTypeSerializer<?>> getTypeSerializers() {
		return typeSerializers;
	}
	
	public ExceptionPolicy getExceptionPolicy() {
		return exceptionPolicy;
	}
	
	public List<JsonSerializeApproach> getSerializeApproaches() {
		return serializeApproaches;
	}
	
	public void addSerializeApproach(JsonSerializeApproach serializeApproach) {
		getSerializeApproaches().add(serializeApproach);
	}
	
	public void addTypeSerializer(JsonTypeSerializer<?> typeSerializer) {
		getTypeSerializers().put(typeSerializer.getSerializedType(), typeSerializer);
	}
	
	protected JsonTypeSerializer<?> findTypeSerializer(Class<?> type) {
		return getTypeSerializers().get(type);
	}
	
	public boolean hasTypeSerializer(Class<?> type) {
		return findTypeSerializer(type) != null;
	}
	
	public JsonTypeSerializer<?> getTypeSerializer(Class<?> type) {
		JsonTypeSerializer<?> result=findTypeSerializer(type);
		if(result == null)
			throw new JsonSerializationException("No serializer registered for type: "+type.getName());
		return result;
	}
	
	public void serializeFragment(JsonFragmentBuilder builder, Object value) throws IOException {
		boolean succeeded=false;
		for(JsonSerializeApproach serializationApproach : getSerializeApproaches())
			if(serializationApproach.serialize(this, builder, value)) {
				succeeded = true;
				break;
			}
		if(!succeeded)
			throw new JsonSerializationException("Could not serialize value to JSON fragment: "+value);
	}

	public <T> T serialize(JsonDocumentBuilder<T> builder, Object value) throws IOException {
		serializeFragment(builder, value);
		return builder.end();
	}
}
