/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.serialization.serialize.approach;

import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import com.sigpwned.jsonification.JsonSerializer;
import com.sigpwned.jsonification.manipulation.JsonFragmentBuilder;
import com.sigpwned.jsonification.manipulation.build.BufferedJsonFragmentBuilder;
import com.sigpwned.jsonification.serialization.JsonSerializationException;
import com.sigpwned.jsonification.serialization.serialize.JsonSerializeApproach;

public class BeanJsonSerializeApproach extends JsonSerializeApproach {
	@Override
	public boolean serialize(JsonSerializer serializer, JsonFragmentBuilder builder, Object value) throws IOException {
		boolean result;
		
		if(value != null) {
			BufferedJsonFragmentBuilder builder2=new BufferedJsonFragmentBuilder();
			result = value(serializer, builder2, value);
			if(result)
				builder2.replay(builder);
		}
		else {
			builder.nullValue();
			result = true;
		}
		
		return result;
	}
	
	protected boolean value(JsonSerializer serializer, JsonFragmentBuilder builder, Object value) throws IOException {
		boolean result;
		
		if(value == null) {
			builder.nullValue();
			result = true;
		} else
		if(value instanceof Byte || value instanceof Short || value instanceof Integer || value instanceof Long) {
			Number number=(Number) value;
			builder.value(number.longValue());
			result = true;
		} else
		if(value instanceof Float || value instanceof Double) {
			Number number=(Number) value;
			builder.value(number.doubleValue());
			result = true;
		} else
		if(value instanceof Boolean) {
			builder.value((Boolean) value);
			result = true;
		} else
		if(value instanceof String) {
			builder.value((String) value);
			result = true;
		} else
		if(value instanceof Collection<?>) {
			Collection<?> values=(Collection<?>) value;
			builder.beginList();
			try {
				int index=0;
				Iterator<?> iter=values.iterator();
				while(iter.hasNext()) {
					Object element=iter.next();
					serializer.serializeFragment(builder, element);
					index = index+1;
				}
			}
			finally {
				builder.endList();
			}
			result = true;
		}
		else {
			builder.beginObject();
			try {
				int succeeded=0;
				
				List<Getter> getters=getters(value.getClass());
				for(Getter getter : getters) {
					Val val=getter.get(serializer, value);
					if(val != null) {
						builder.name(getter.getName());
						serializer.serializeFragment(builder, val.getValue());
						succeeded = succeeded+1;
					}
					else {
						// We couldn't retrieve this value. Ignore it.
					}
				}
				
				result = (succeeded != 0);
			}
			finally {
				builder.endObject();
			}
		}
		return result;
	}
	
	protected class Val {
		public Object value;
		
		public Val(Object value) {
			this.value = value;
		}
		
		public Object getValue() {
			return value;
		}
	}

	protected interface Getter {
		public String getName();
		public Val get(JsonSerializer serializer, Object o);
	}
	
	protected class FieldGetter implements Getter {
		private Field field;
		
		public FieldGetter(Field field) {
			this.field = field;
		}
		
		public Field getField() {
			return field;
		}
		
		public String getName() {
			return field.getName();
		}
		
		public Val get(JsonSerializer serializer, Object o) {
			Val result;
			try {
				result = new Val(getField().get(o));
			}
			catch(IllegalArgumentException e) {
				JsonSerializationException e2=serializer.getExceptionPolicy().format(e);
				if(e2 != null)
					throw e2;
				result = null;
			}
			catch(IllegalAccessException e) {
				JsonSerializationException e2=serializer.getExceptionPolicy().format(e);
				if(e2 != null)
					throw e2;
				result = null;
			}
			return result;
		}
	}
	
	protected class MethodGetter implements Getter {
		private Field field;
		private Method getter;
		
		public MethodGetter(Field field, Method getter) {
			this.field = field;
			this.getter = getter;
		}
		
		public Field getField() {
			return field;
		}
		
		public String getName() {
			return getField().getName();
		}

		public Method getGetter() {
			return getter;
		}
		
		public Val get(JsonSerializer serializer, Object o) {
			Val result;
			try {
				result = new Val(getGetter().invoke(o));
			}
			catch(IllegalArgumentException e) {
				JsonSerializationException e2=serializer.getExceptionPolicy().format(e);
				if(e2 != null)
					throw e2;
				result = null;
			}
			catch(IllegalAccessException e) {
				JsonSerializationException e2=serializer.getExceptionPolicy().format(e);
				if(e2 != null)
					throw e2;
				result = null;
			}
			catch(InvocationTargetException e) {
				JsonSerializationException e2=serializer.getExceptionPolicy().format(e);
				if(e2 != null)
					throw e2;
				result = null;
			}
			return result;
		}
	}
	
	protected List<Getter> getters(Class<?> type) {
		List<Getter> result=new ArrayList<Getter>();
		for(Class<?> c=type;c!=null;c=c.getSuperclass())
			for(Field field : c.getDeclaredFields())
				if((field.getModifiers()&Modifier.STATIC) != 0) {
					// We never serialize static fields
				} else
				if((field.getModifiers()&Modifier.PUBLIC) != 0)
					result.add(new FieldGetter(field));
				else {
					List<String> names=getterNames(field);
	
					Method method=findMethod(type, field.getType(), names);
					
					if(method != null)
						result.add(new MethodGetter(field, method));
				}
		return result;
	}
	
	protected List<String> getterNames(Field field) {
		List<String> result=new ArrayList<String>();
		
		String slug=capitalizeFirstLetter(field.getName());
		if(field.getType().equals(boolean.class) || field.getType().equals(Boolean.class));
			result.add("is"+slug);
		result.add("get"+slug);
		
		return result;
	}
}

