/*
 * Copyright 2013 Andy Boothe
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package com.sigpwned.jsonification.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;

import com.sigpwned.jsonification.x.JsonParseException;

public abstract class Parser<T> {
	public static enum Cleanup {
		NOP {
			@Override
			public void cleanup(Reader inputSource) throws IOException {
			}
		},
		CLOSE {
			@Override
			public void cleanup(Reader inputSource) throws IOException {
				inputSource.close();
			}
		},
		DRAIN {
			@Override
			public void cleanup(Reader inputSource) throws IOException {
				try {
					for(int ch=inputSource.read();ch!=-1;ch=inputSource.read())
						if(!Character.isWhitespace(ch))
							throw new JsonParseException("Input left after JSON value");
				}
				finally {
					inputSource.close();
				}
			}
		};
		
		public abstract void cleanup(Reader inputSource) throws IOException;
		
		public static Cleanup defaultValue() {
			return CLOSE;
		}
	}

	private Reader inputSource;

	private Reader getInputSource() {
		return inputSource;
	}
	
	protected void setInputSource(Reader inputSource) {
		if(inputSource != null) {
			if(!inputSource.markSupported())
				inputSource = new BufferedReader(inputSource);
			this.inputSource = inputSource;
		}
		else
			this.inputSource = null;
	}
	
	public T parse(InputStream input) throws IOException {
		return parse(new InputStreamReader(input));
	}
	
	public T parse(Reader input) throws IOException {
		return parse(input, Cleanup.defaultValue());
	}
	
	public T parse(String input) {
	    T result;
	    try {
	        result = parse(new StringReader(input), Cleanup.DRAIN);
	    }
	    catch(IOException e) {
	        // This should never happen. It's a StringReader.
	        throw new RuntimeException(e);
	    }
	    return result;
	}
	
	public synchronized T parse(Reader inputSource, Cleanup cleanup) throws IOException {
		T result;
		
		setInputSource(inputSource);
		try {
			try {
				result = parse();
			}
			finally {
				cleanup.cleanup(getInputSource());
			}
		}
		finally {
			setInputSource(null);
		}
		
		return result;
	}

	protected abstract T parse() throws IOException;
	
	protected void whitespace() throws IOException {
		while(Character.isWhitespace(peek()))
			read();
	}
	
	protected int read() throws IOException {
		return getInputSource().read();
	}
	
	protected int peek() throws IOException {
		getInputSource().mark(1);
		
		int result;
		try {
			result = getInputSource().read();
		}
		finally {
			getInputSource().reset();
		}
		
		return result;
	}
	
	protected void consumei(int want) throws IOException {
		int got=read();
		if(Character.toLowerCase(got) != Character.toLowerCase(want))
			throw parseException("Expected `"+((char)want)+"', got `"+((char)got)+"'");
	}
	
	protected void consume(int want) throws IOException {
		int got=read();
		if(got != want)
			throw parseException("Expected `"+((char)want)+"', got `"+((char)got)+"'");
	}
	
	protected boolean attempti(int ch) throws IOException {
		ch = Character.toLowerCase(ch);
		boolean result;
		if(Character.toLowerCase(peek()) == ch) {
			read();
			result = true;
		}
		else
			result = false;
		return result;
	}
	
	protected boolean attempt(int ch) throws IOException {
		boolean result;
		if(peek() == ch) {
			read();
			result = true;
		}
		else
			result = false;
		return result;
	}
	
	protected JsonParseException parseException(String message) {
		throw new JsonParseException(message);
	}
}
